import abc
import logging
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import BaseModel, PrivateAttr


class OperatorRegistry(type(BaseModel)):
    _registry = {}
    _parametrized_registry = []

    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        # Only register if it's a concrete class that inherits from Operand
        base_classes = [b for b in bases]
        base_names = {b.__name__ for b in bases}
        while base_classes:
            base = base_classes.pop()
            base_names.update(b.__name__ for b in base.__bases__)
            base_classes.extend(base.__bases__)

        if "Operator" in base_names:
            # Track parametrized operands separately
            if "ParametrizedOperator" in base_names:
                cls._parametrized_registry.append(new_class)
            else:
                try:
                    instance = new_class()
                    cls._registry[instance.name] = new_class
                    if instance.alias:
                        cls._registry[instance.alias] = new_class
                except Exception:
                    pass
        return new_class

    @classmethod
    def get_operator(cls, name: str) -> Optional["Operator"]:
        # First try to resolve as a parametrized operand formula
        for operator_cls in cls._parametrized_registry:
            resolved = operator_cls.from_formula(name)
            if resolved is not None:
                return resolved
        # Fall back to direct registry lookup
        non_parametrized = cls._registry.get(name)
        if non_parametrized is not None:
            return non_parametrized()
        return None


class Operator(BaseModel, metaclass=OperatorRegistry):
    name: Optional[str] = None
    alias: Optional[str] = None
    is_unary: bool = False
    is_symmetrical: bool = False
    has_symmetry_importance: bool = False
    input_type: Optional[str] = None
    output_type: Optional[str] = None
    is_categorical: bool = False
    is_vectorizable: bool = False
    is_grouping: bool = False
    is_binary: bool = False
    is_vector: bool = False
    is_distribution_dependent: bool = False
    params: Optional[Dict[str, str]] = None

    _logger: logging.Logger = PrivateAttr(default=logging.getLogger(__name__))

    def set_params(self, params: Dict[str, str]):
        self.params = params
        return self

    def get_params(self) -> Dict[str, Optional[str]]:
        res = {"alias": self.alias}
        res.update(self.params or {})
        return res

    def to_formula(self) -> str:
        return self.name

    def get_hash_component(self) -> str:
        return self.to_formula()

    def set_logger(self, logger: logging.Logger):
        self._logger = logger
        return self

    def delete_data(self):
        pass

    def rename_params(self, columns_renaming: Dict[str, str]) -> "Operator":
        # Rename occurrences of column names inside self.params keys according to columns_renaming
        if not self.params or not columns_renaming:
            return self

        # Replace longer keys first to avoid partial overlaps
        replacements = sorted(columns_renaming.items(), key=lambda kv: -len(kv[0]))

        renamed_params: Dict[str, str] = {}
        for param_key, param_value in self.params.items():
            new_key = param_key
            for old, new in replacements:
                if old and old in new_key:
                    new_key = new_key.replace(old, new)

            if new_key in renamed_params and new_key != param_key:
                self._logger.warning(
                    "Param key collision after rename: '%s' -> '%s'. Overwriting value.",
                    param_key,
                    new_key,
                )
            renamed_params[new_key] = param_value

        self.params = renamed_params
        return self


class ParametrizedOperator(Operator, abc.ABC):

    @abc.abstractmethod
    def to_formula(self) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def from_formula(cls, formula: str) -> Optional["Operator"]:
        pass


MAIN_COLUMN = "main_column"


class PandasOperator(Operator, abc.ABC):
    def calculate(self, **kwargs) -> pd.Series:
        if self.is_unary:
            return self.calculate_unary(kwargs["data"])
        elif self.is_binary or self.is_grouping:
            return self.calculate_binary(kwargs["left"], kwargs["right"])
        else:
            return self.calculate_vector(kwargs["data"])

    def calculate_unary(self, data: pd.Series) -> pd.Series:
        pass

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        pass

    def calculate_vector(self, data: List[pd.Series]) -> pd.Series:
        pass

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        if not self.is_vectorizable:
            raise RuntimeError(f"Cannot apply calculate_group: operator {self.name} is not vectorizable")
        else:
            raise RuntimeError(f"Unimplemented calculate_group for operator {self.name}")

    def _loc(self, df_to, df_from):
        df_from.loc[np.nan] = np.nan
        return df_to.fillna(np.nan).apply(lambda x: df_from.loc[x])

    def _round_value(
        self, value: Union[pd.Series, pd.DataFrame], precision: Optional[int] = None
    ) -> Union[pd.Series, pd.DataFrame]:
        if isinstance(value, pd.DataFrame):
            return value.apply(self._round_value, axis=1)

        if np.issubdtype(value.dtype, np.floating):
            precision = precision or np.finfo(value.dtype).precision
            return np.trunc(value * 10**precision) / (10**precision)
        else:
            return value


class VectorizableMixin(Operator):
    group_index: int = 1

    def validate_calculation(self, input_columns: List[str], **kwargs) -> Tuple[str, List[str]]:
        if not kwargs.get(MAIN_COLUMN):
            raise ValueError(f"Expected argument {MAIN_COLUMN} for grouping operator {self.name} not found")
        group_column = kwargs[MAIN_COLUMN]
        value_columns = [col for col in input_columns if col != group_column]

        return group_column, value_columns



import hashlib
from logging import Logger, getLogger
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_bool_dtype as is_bool
from pandas.api.types import (
    is_float_dtype,
    is_numeric_dtype,
    is_object_dtype,
    is_string_dtype,
)

from upgini.errors import ValidationError
from upgini.metadata import (
    ENTITY_SYSTEM_RECORD_ID,
    EVAL_SET_INDEX,
    SEARCH_KEY_UNNEST,
    SYSTEM_RECORD_ID,
    TARGET,
    SearchKey,
)
from upgini.resource_bundle import ResourceBundle, get_custom_bundle
from upgini.utils import find_numbers_with_decimal_comma
from upgini.utils.country_utils import CountrySearchKeyConverter
from upgini.utils.datetime_utils import DateTimeConverter
from upgini.utils.ip_utils import IpSearchKeyConverter
from upgini.utils.one_hot_encoder import OneHotDecoder
from upgini.utils.phone_utils import PhoneSearchKeyConverter
from upgini.utils.postal_code_utils import PostalCodeSearchKeyConverter


class Normalizer:

    MAX_STRING_FEATURE_LENGTH = 24573

    def __init__(
        self,
        bundle: ResourceBundle = None,
        logger: Logger = None,
    ):
        self.bundle = bundle or get_custom_bundle()
        self.logger = logger or getLogger()
        self.columns_renaming = {}
        self.search_keys = {}
        self.generated_features = []
        self.removed_datetime_features = []
        self.true_one_hot_groups: dict[str, list[str]] | None = None
        self.pseudo_one_hot_groups: dict[str, list[str]] | None = None

    def normalize(
        self, df: pd.DataFrame, search_keys: Dict[str, SearchKey], generated_features: List[str]
    ) -> Tuple[pd.DataFrame, Dict[str, SearchKey], List[str]]:
        self.search_keys = search_keys.copy()
        self.generated_features = generated_features.copy()

        df = df.copy()

        df = self._convert_one_hot_encoded_columns(df)

        df = self._rename_columns(df)

        df = self._remove_dates_from_features(df)

        df = self._cut_too_long_string_values(df)

        df = self._convert_bools(df)

        df = self._convert_float16(df)

        df = self._correct_decimal_comma(df)

        df = self._convert_phone_numbers(df)

        df = self._convert_ip_addresses(df)

        df = self._convert_postal_codes(df)

        df = self._convert_countries(df)

        df = self.__convert_features_types(df)

        return df, self.search_keys, self.generated_features

    def _convert_one_hot_encoded_columns(self, df: pd.DataFrame):
        if self.true_one_hot_groups is not None or self.pseudo_one_hot_groups is not None:
            df = OneHotDecoder.decode_with_cached_groups(
                df, self.true_one_hot_groups, self.pseudo_one_hot_groups
            )
        else:
            df, self.true_one_hot_groups, self.pseudo_one_hot_groups = OneHotDecoder.decode(df)
        return df

    def _rename_columns(self, df: pd.DataFrame):
        # logger.info("Replace restricted symbols in column names")
        new_columns = []
        dup_counter = 0
        for column in df.columns:
            if column in [
                TARGET,
                EVAL_SET_INDEX,
                SYSTEM_RECORD_ID,
                ENTITY_SYSTEM_RECORD_ID,
                SEARCH_KEY_UNNEST,
                DateTimeConverter.DATETIME_COL,
            ]:
                self.columns_renaming[column] = column
                new_columns.append(column)
                continue

            new_column = add_hash_suffix(column, self.bundle)

            # if column starts with non alphabetic symbol then add "a" to the beginning of string
            if ord(new_column[0]) not in range(ord("a"), ord("z") + 1):
                new_column = "a" + new_column

            # replace unsupported characters to "_"
            for idx, c in enumerate(new_column):
                if ord(c) not in range(ord("a"), ord("z") + 1) and ord(c) not in range(ord("0"), ord("9") + 1):
                    new_column = new_column[:idx] + "_" + new_column[idx + 1 :]

            if new_column in new_columns:
                new_column = f"{new_column}_{dup_counter}"
                dup_counter += 1
            new_columns.append(new_column)

            # df.columns.values[col_idx] = new_column
            # rename(columns={column: new_column}, inplace=True)

            if new_column != column:
                if column in self.search_keys:
                    self.search_keys[new_column] = self.search_keys[column]
                    del self.search_keys[column]
                if column in self.generated_features:
                    self.generated_features.remove(column)
                    self.generated_features.append(new_column)
            self.columns_renaming[new_column] = str(column)
        df.columns = new_columns
        return df

    def _get_features(self, df: pd.DataFrame) -> List[str]:
        system_columns = [ENTITY_SYSTEM_RECORD_ID, EVAL_SET_INDEX, SEARCH_KEY_UNNEST, SYSTEM_RECORD_ID, TARGET]
        features = set(df.columns) - set(self.search_keys.keys()) - set(system_columns)
        return sorted(list(features))

    def _remove_dates_from_features(self, df: pd.DataFrame):
        features = self._get_features(df)

        for f in features:
            converter = DateTimeConverter(f)
            if converter.is_datetime(df) and f != DateTimeConverter.DATETIME_COL:
                self.removed_datetime_features.append(f)
                df.drop(columns=f, inplace=True)

        return df

    def _cut_too_long_string_values(self, df: pd.DataFrame):
        """Check that string values less than maximum characters for LLM"""
        # logger.info("Validate too long string values")
        for col in df.columns:
            if is_string_dtype(df[col]) or is_object_dtype(df[col]):
                max_length: int = df[col].astype("str").str.len().max()
                if max_length > self.MAX_STRING_FEATURE_LENGTH:
                    df[col] = df[col].astype("str").str.slice(stop=self.MAX_STRING_FEATURE_LENGTH)

        return df

    @staticmethod
    def _convert_bools(df: pd.DataFrame):
        """Convert bool columns to string"""
        # logger.info("Converting bool to int")
        for col in df.columns:
            if is_bool(df[col]):
                df[col] = df[col].astype("str")
        return df

    @staticmethod
    def _convert_float16(df: pd.DataFrame):
        """Convert float16 to float"""
        # logger.info("Converting float16 to float")
        for col in df.columns:
            if is_float_dtype(df[col]):
                df[col] = df[col].astype("float64")
        return df

    def _correct_decimal_comma(self, df: pd.DataFrame):
        """Check DataSet for decimal commas and fix them"""
        # logger.info("Correct decimal commas")
        columns_to_fix = find_numbers_with_decimal_comma(df)
        if len(columns_to_fix) > 0:
            self.logger.warning(f"Convert strings with decimal comma to float: {columns_to_fix}")
            for col in columns_to_fix:
                df[col] = df[col].astype("string").str.replace(",", ".", regex=False).astype(np.float64)
        return df

    def _convert_phone_numbers(self, df: pd.DataFrame) -> pd.DataFrame:
        maybe_country_col = SearchKey.find_key(self.search_keys, SearchKey.COUNTRY)
        for phone_col in SearchKey.find_all_keys(self.search_keys, SearchKey.PHONE):
            converter = PhoneSearchKeyConverter(phone_col, maybe_country_col)
            df = converter.convert(df)
        return df

    def _convert_ip_addresses(self, df: pd.DataFrame) -> pd.DataFrame:
        for ip_col in SearchKey.find_all_keys(self.search_keys, SearchKey.IP):
            df[ip_col] = df[ip_col].apply(IpSearchKeyConverter.safe_ip_parse)
        return df

    def _convert_postal_codes(self, df: pd.DataFrame) -> pd.DataFrame:
        for postal_code_col in SearchKey.find_all_keys(self.search_keys, SearchKey.POSTAL_CODE):
            df = PostalCodeSearchKeyConverter(postal_code_col).convert(df)
        return df

    def _convert_countries(self, df: pd.DataFrame) -> pd.DataFrame:
        maybe_country_col = SearchKey.find_key(self.search_keys, SearchKey.COUNTRY)
        if maybe_country_col:
            df = CountrySearchKeyConverter(maybe_country_col).convert(df)
        return df

    def __convert_features_types(self, df: pd.DataFrame):
        # self.logger.info("Convert features to supported data types")

        for f in self._get_features(df):
            if not is_numeric_dtype(df[f]):
                df[f] = df[f].astype("string")
        return df


def add_hash_suffix(column: str, bundle: ResourceBundle | None = None) -> str:
    new_column = str(column)
    suffix = hashlib.sha256(new_column.encode()).hexdigest()[:6]
    if bundle is not None and len(new_column) == 0:
        raise ValidationError(bundle.get("dataset_empty_column_names"))
    # db limit for column length
    if len(new_column) > 250:
        new_column = new_column[:250]

    # make column name unique relative to server features
    new_column = f"{new_column}_{suffix}"

    new_column = new_column.lower()
    return new_column



import pandas as pd
from pandas.api.types import is_object_dtype, is_string_dtype

from upgini.utils.base_search_key_detector import BaseSearchKeyDetector


class CountrySearchKeyConverter:

    def __init__(self, country_col: str):
        self.country_col = country_col

    def convert(self, df: pd.DataFrame) -> pd.DataFrame:
        df[self.country_col] = (
            df[self.country_col]
            .astype("string")
            .str.upper()
            .str.replace(r"[^A-Z]", "", regex=True)
            .str.replace("UK", "GB", regex=False)
        )
        return df


class CountrySearchKeyDetector(BaseSearchKeyDetector):
    def _is_search_key_by_name(self, column_name: str) -> bool:
        return "country" in str(column_name).lower()

    def _is_search_key_by_values(self, column: pd.Series) -> bool:
        if not is_string_dtype(column) and not is_object_dtype(column):
            return False

        all_count = len(column)
        is_countries_count = len(column[column.astype("string").str.upper().isin(self.COUNTRY_CODES)])
        return is_countries_count / all_count > 0.1

    @staticmethod
    def convert_country_to_iso_code(df: pd.DataFrame, country_column: str) -> pd.DataFrame:
        if df[country_column].isin(CountrySearchKeyDetector.COUNTRIES.values()).all():
            return df

        df[country_column] = (
            df[country_column]
            .astype("string")
            .str.upper()
            .map(CountrySearchKeyDetector.COUNTRIES)
            .fillna(df[country_column])
        )

        return df

    COUNTRIES = {
        "BOLIVIA": "BO",
        "ALBANIA": "AL",
        "ALGERIA": "DZ",
        "ANDORRA": "AD",
        "ANGOLA": "AO",
        "ARGENTINA": "AR",
        "ARMENIA": "AM",
        "ARUBA": "AW",
        "AUSTRALIA": "AU",
        "AUSTRIA": "AT",
        "AZERBAIJAN": "AZ",
        "BAHAMAS": "BS",
        "BAHRAIN": "BH",
        "BANGLADESH": "BD",
        "BARBADOS": "BB",
        "BELARUS": "BY",
        "BELGIUM": "BE",
        "BELIZE": "BZ",
        "BENIN": "BJ",
        "BERMUDA": "BM",
        "BHUTAN": "BT",
        "BOSNIA AND HERZEGOVINA": "BA",
        "BOTSWANA": "BW",
        "BRAZIL": "BR",
        "BRUNEI DARUSSALAM": "BN",
        "BRUNEI": "BN",
        "BULGARIA": "BG",
        "BURKINA FASO": "BF",
        "BURMA": "MM",
        "BURUNDI": "BI",
        "CABO VERDE": "CV",
        "CAMBODIA": "KH",
        "CAMEROON": "CM",
        "CANADA": "CA",
        "CENTRAL AFRICAN REPUBLIC": "CF",
        "CHAD": "TD",
        "CHILE": "CL",
        "CHINA": "CN",
        "COLOMBIA": "CO",
        "CONGO": "CG",
        "COSTA RICA": "CR",
        "COTE D'IVOIRE": "CI",
        "CROATIA": "HR",
        "CUBA": "CU",
        "CYPRUS": "CY",
        "CZECHIA": "CZ",
        "CZECH REPUBLIC": "CZ",
        "DENMARK": "DK",
        "DJIBOUTI": "DJ",
        "DOMINICA": "DM",
        "DOMINICAN REPUBLIC": "DO",
        "ECUADOR": "EC",
        "EGYPT": "EG",
        "EL SALVADOR": "SV",
        "ENGLAND": "GB",
        "EQUATORIAL GUINEA": "GQ",
        "ERITREA": "ER",
        "ESTONIA": "EE",
        "ESWATINI": "SZ",
        "ETHIOPIA": "ET",
        "FIJI": "FJ",
        "FINLAND": "FI",
        "FRANCE": "FR",
        "GABON": "GA",
        "GAMBIA": "GM",
        "GEORGIA": "GE",
        "GERMANY": "DE",
        "GHANA": "GH",
        "GIBRALTAR": "GI",
        "GREAT BRITAIN": "GB",
        "GREECE": "GR",
        "GRENADA": "GD",
        "GUADELOUPE": "GP",
        "GUATEMALA": "GT",
        "GUINEA": "GN",
        "GUINEA-BISSAU": "GW",
        "GUYANA": "GY",
        "HAITI": "HT",
        "HOLY SEE": "VA",
        "HONDURAS": "HN",
        "HONG KONG": "HK",
        "HUNGARY": "HU",
        "ICELAND": "IS",
        "INDIA": "IN",
        "INDONESIA": "ID",
        "IRAN": "IR",
        "IRAQ": "IQ",
        "IRELAND": "IE",
        "ISRAEL": "IL",
        "ITALY": "IT",
        "JAMAICA": "JM",
        "JAPAN": "JP",
        "JORDAN": "JO",
        "KAZAKHSTAN": "KZ",
        "KENYA": "KE",
        "DPRK": "KP",
        "REPUBLIC OF KOREA": "KR",
        "KUWAIT": "KW",
        "KYRGYZSTAN": "KG",
        "LAOS": "LA",
        "LATVIA": "LV",
        "LEBANON": "LB",
        "LESOTHO": "LS",
        "LIBERIA": "LR",
        "LIBYA": "LY",
        "LIECHTENSTEIN": "LI",
        "LITHUANIA": "LT",
        "LUXEMBOURG": "LU",
        "NORTH MACEDONIA": "MK",
        "MADAGASCAR": "MG",
        "MALAWI": "MW",
        "MALAYSIA": "MY",
        "MALDIVES": "MV",
        "MALI": "ML",
        "MALTA": "MT",
        "MAURITANIA": "MR",
        "MEXICO": "MX",
        "MOLDOVA": "MD",
        "MONACO": "MC",
        "MONGOLIA": "MN",
        "MONTENEGRO": "ME",
        "MOROCCO": "MA",
        "MOZAMBIQUE": "MZ",
        "MYANMAR": "MM",
        "NAMIBIA": "NA",
        "NEPAL": "NP",
        "NETHERLANDS": "NL",
        "NEW ZEALAND": "NZ",
        "NICARAGUA": "NI",
        "NIGER": "NE",
        "NIGERIA": "NG",
        "NORTH KOREA": "KP",
        "NORTHERN IRELAND": "GB",
        "NORWAY": "NO",
        "OMAN": "OM",
        "PAKISTAN": "PK",
        "PALESTINE": "PS",
        "PANAMA": "PA",
        "PAPUA NEW GUINEA": "PG",
        "PARAGUAY": "PY",
        "PERU": "PE",
        "PHILIPPINES": "PH",
        "POLAND": "PL",
        "PORTUGAL": "PT",
        "PUERTO RICO": "PR",
        "QATAR": "QA",
        "ROMANIA": "RO",
        "RUSSIAN FEDERATION": "RU",
        "RUSSIA": "RU",
        "RWANDA": "RW",
        "SAHRAWI ARAB DEMOCRATIC REPUBLIC": "EH",
        "SAN MARINO": "SM",
        "SAUDI ARABIA": "SA",
        "SCOTLAND": "GB",
        "SENEGAL": "SN",
        "SERBIA": "RS",
        "SIERRA LEONE": "SL",
        "SINGAPORE": "SG",
        "SINT MAARTEN": "SX",
        "SLOVAKIA": "SK",
        "SLOVENIA": "SI",
        "SOMALIA": "SO",
        "SOUTH AFRICA": "ZA",
        "SOUTH KOREA": "KR",
        "SOUTH SUDAN": "SS",
        "SPAIN": "ES",
        "SRI LANKA": "LK",
        "SUDAN": "SD",
        "SURINAME": "SR",
        "SWEDEN": "SE",
        "SWITZERLAND": "CH",
        "SYRIA": "SY",
        "TAIWAN": "TW",
        "TAJIKISTAN": "TJ",
        "TANZANIA": "TZ",
        "THAILAND": "TH",
        "TOGO": "TG",
        "TUNISIA": "TN",
        "TURKEY": "TR",
        "TURKMENISTAN": "TM",
        "UGANDA": "UG",
        "UKRAINE": "UA",
        "UNITED ARAB EMIRATES": "AE",
        "UNITED KINGDOM OF GREAT BRITAIN AND NORTHERN IRELAND": "GB",
        "UNITED STATES OF AMERICA": "US",
        "UNITED STATES": "US",
        "USA": "US",
        "URUGUAY": "UY",
        "UZBEKISTAN": "UZ",
        "VANUATU": "VU",
        "VATICAN": "VA",
        "VENEZUELA": "VE",
        "VIETNAM": "VN",
        "WALES": "GB",
        "WESTERN SAHARA": "EH",
        "YEMEN": "YE",
        "ZAMBIA": "ZM",
        "ZIMBABWE": "ZW",
    }

    COUNTRY_CODES = list(COUNTRIES.keys()) + list(set(COUNTRIES.values())) + ["Unknown code"]



"""Class to perform random under-sampling."""

# Authors: Guillaume Lemaitre <g.lemaitre58@gmail.com>
#          Christos Aridas
# License: MIT

import numpy as np
from sklearn.utils import _safe_indexing, check_random_state

from .base import BaseUnderSampler
from .utils import _deprecate_positional_args, check_target_type


class RandomUnderSampler(BaseUnderSampler):
    """Class to perform random under-sampling.

    Under-sample the majority class(es) by randomly picking samples
    with or without replacement.

    Read more in the :ref:`User Guide <controlled_under_sampling>`.

    Parameters
    ----------
    {sampling_strategy}

    {random_state}

    replacement : bool, default=False
        Whether the sample is with or without replacement.

    Attributes
    ----------
    sampling_strategy_ : dict
        Dictionary containing the information to sample the dataset. The keys
        corresponds to the class labels from which to sample and the values
        are the number of samples to sample.

    sample_indices_ : ndarray of shape (n_new_samples,)
        Indices of the samples selected.

        .. versionadded:: 0.4

    n_features_in_ : int
        Number of features in the input dataset.

        .. versionadded:: 0.9

    See Also
    --------
    NearMiss : Undersample using near-miss samples.

    Notes
    -----
    Supports multi-class resampling by sampling each class independently.
    Supports heterogeneous data as object array containing string and numeric
    data.

    Examples
    --------
    >>> from collections import Counter
    >>> from sklearn.datasets import make_classification
    >>> from imblearn.under_sampling import \
RandomUnderSampler # doctest: +NORMALIZE_WHITESPACE
    >>> X, y = make_classification(n_classes=2, class_sep=2,
    ...  weights=[0.1, 0.9], n_informative=3, n_redundant=1, flip_y=0,
    ... n_features=20, n_clusters_per_class=1, n_samples=1000, random_state=10)
    >>> print('Original dataset shape %s' % Counter(y))
    Original dataset shape Counter({{1: 900, 0: 100}})
    >>> rus = RandomUnderSampler(random_state=42)
    >>> X_res, y_res = rus.fit_resample(X, y)
    >>> print('Resampled dataset shape %s' % Counter(y_res))
    Resampled dataset shape Counter({{0: 100, 1: 100}})
    """

    @_deprecate_positional_args
    def __init__(self, *, sampling_strategy="auto", random_state=None, replacement=False):
        super().__init__(sampling_strategy=sampling_strategy)
        self.random_state = random_state
        self.replacement = replacement

    def _check_X_y(self, X, y):
        y, binarize_y = check_target_type(y, indicate_one_vs_all=True)
        try:
            X, y = self._validate_data(
                X,
                y,
                reset=True,
                accept_sparse=["csr", "csc"],
                dtype=None,
                force_all_finite=False,
            )
        except AttributeError:
            from sklearn.utils.validation import check_X_y
            X, y = check_X_y(
                X,
                y,
                accept_sparse=["csr", "csc"],
                dtype=None,
                ensure_all_finite=False,
            )
        return X, y, binarize_y

    def _fit_resample(self, X, y):
        random_state = check_random_state(self.random_state)

        idx_under = np.empty((0,), dtype=int)

        for target_class in np.unique(y):
            if target_class in self.sampling_strategy_.keys():
                n_samples = self.sampling_strategy_[target_class]
                index_target_class = random_state.choice(
                    range(np.count_nonzero(y == target_class)),
                    size=n_samples,
                    replace=self.replacement,
                )
            else:
                index_target_class = slice(None)

            idx_under = np.concatenate(
                (
                    idx_under,
                    np.flatnonzero(y == target_class)[index_target_class],
                ),
                axis=0,
            )

        self.sample_indices_ = idx_under

        return _safe_indexing(X, idx_under), _safe_indexing(y, idx_under)

    def _more_tags(self):
        return {
            "X_types": ["2darray", "string", "sparse", "dataframe"],
            "sample_indices": True,
            "allow_nan": True,
        }



import numpy as np
import pandas as pd


class OneHotDecoder:

    def encode(df: pd.DataFrame, category_columns: list[str]) -> pd.DataFrame:
        """
        Encode categorical columns into one-hot encoded columns.
        """
        return pd.get_dummies(df, columns=category_columns, prefix_sep="")

    def decode(df: pd.DataFrame) -> (pd.DataFrame, dict[str, list[str]], dict[str, list[str]]):
        """
        Detect one-hot encoded column groups and collapse each group into a single
        categorical column. For each row, all active bits in the group are
        encoded into a unique category using a bitmask over the group's columns
        (ordered by numeric suffix). Rows with zero active bits are set to NA.

        Returns a new DataFrame with transformed columns.
        """
        one_hot_candidate_groups = OneHotDecoder._group_one_hot_fast(df.columns)
        true_one_hot_groups: dict[str, list[str]] = {}

        # 1) Detect valid one-hot groups (filter candidates by column-level checks)
        for group_name, column_candidates in one_hot_candidate_groups.items():
            group_columns: list[str] = []
            for column in column_candidates:
                value_counts = df[column].value_counts(dropna=False, normalize=True)
                most_frequent_percent = value_counts.iloc[0]
                if most_frequent_percent >= 0.6 and OneHotDecoder._is_one_hot_encoded(df[column]):
                    group_columns.append(column)
            if len(group_columns) > 1:
                true_one_hot_groups[group_name] = group_columns

        # 2) Transform: replace each detected group with one categorical column
        if not true_one_hot_groups:
            return df, {}, {}

        result_df = df.copy()
        pseudo_one_hot_groups: dict[str, list[str]] = {}
        for group_name, group_columns in true_one_hot_groups.items():
            sub = result_df[group_columns].copy()
            for c in group_columns:
                s = sub[c]
                if s.dtype == "object" or s.dtype == "string":
                    s = s.astype(str).str.strip().str.lower()
                    s = s.replace({"true": "1", "false": "0"})
                s = pd.to_numeric(s, errors="coerce")
                sub[c] = s

            # 3) Find pseudo one-hot encoded columns when there are multiple ones in one row
            if any(sub.sum(axis=1) > 1):
                pseudo_one_hot_groups[group_name] = group_columns
                result_df[group_columns] = result_df[group_columns].astype("string")
                continue

            # Coerce values to numeric 0/1 handling common textual forms
            sub = sub.fillna(0.0)
            # Binarize strictly to 0/1
            bin_values = (sub.to_numpy() > 0.5).astype(np.int64)
            # Map single active bit to exact numeric suffix from column name
            row_sums = bin_values.sum(axis=1)
            argmax_idx = bin_values.argmax(axis=1)
            suffix_arr = np.array(
                [int(OneHotDecoder._split_prefix_numeric_suffix(col)[1]) for col in group_columns], dtype=np.int64
            )
            codes = suffix_arr[argmax_idx]
            categorical_series = pd.Series(codes, index=sub.index)
            # Keep only rows with exactly one active bit; else set NA
            categorical_series = categorical_series.where(row_sums == 1, other=pd.NA)
            # Use pandas nullable integer dtype to keep NA with integer codes
            result_df[group_name] = categorical_series.astype("Int64").astype("string")

            # Drop original one-hot columns of the group
            result_df = result_df.drop(columns=group_columns)

        for group_name in pseudo_one_hot_groups:
            del true_one_hot_groups[group_name]

        return result_df, true_one_hot_groups, pseudo_one_hot_groups

    def decode_with_cached_groups(
        df: pd.DataFrame, true_one_hot_groups: dict[str, list[str]], pseudo_one_hot_groups: dict[str, list[str]]
    ) -> pd.DataFrame:
        """
        Decode one-hot encoded columns with cached groups.
        """
        result_df = df.copy()
        # 1. Transform regular one-hot groups back to categorical
        if true_one_hot_groups:
            for group_name, group_columns in true_one_hot_groups.items():
                sub = result_df[group_columns].copy()
                for c in group_columns:
                    s = sub[c]
                    if s.dtype == "object" or s.dtype == "string":
                        s = s.astype(str).str.strip().str.lower()
                        s = s.replace({"true": "1", "false": "0"})
                    s = pd.to_numeric(s, errors="coerce")
                    sub[c] = s
                sub = sub.fillna(0.0)
                bin_values = (sub.to_numpy() > 0.5).astype(np.int64)
                row_sums = bin_values.sum(axis=1)
                argmax_idx = bin_values.argmax(axis=1)
                suffix_arr = np.array(
                    [int(OneHotDecoder._split_prefix_numeric_suffix(col)[1]) for col in group_columns], dtype=np.int64
                )
                codes = suffix_arr[argmax_idx]
                categorical_series = pd.Series(codes, index=sub.index)
                categorical_series = categorical_series.where(row_sums == 1, other=pd.NA)
                result_df[group_name] = categorical_series.astype("Int64").astype("string")
                result_df = result_df.drop(columns=group_columns)
        # 2. Convert pseudo-one-hot features to string
        if pseudo_one_hot_groups:
            for _, group_columns in pseudo_one_hot_groups.items():
                result_df[group_columns] = result_df[group_columns].astype("string")
        return result_df

    @staticmethod
    def _is_ascii_digit(c: str) -> bool:
        return "0" <= c <= "9"

    @staticmethod
    def _split_prefix_numeric_suffix(name: str) -> tuple[str, str] | None:
        """
        Return (prefix, numeric_suffix) if name ends with ASCII digits and isn't all digits.
        Otherwise None.
        """
        if not name or not OneHotDecoder._is_ascii_digit(name[-1]):
            return None
        i = len(name) - 1
        while i >= 0 and OneHotDecoder._is_ascii_digit(name[i]):
            i -= 1
        if i < 0:
            # Entire string is digits -> reject
            return None
        return name[: i + 1], name[i + 1 :]  # prefix, suffix

    @staticmethod
    def _group_one_hot_fast(
        candidates: list[str], min_group_size: int = 2, require_consecutive: bool = True
    ) -> dict[str, list[str]]:
        """
        Group OHE-like columns by (prefix, numeric_suffix).
        - Only keeps groups with size >= min_group_size (default: 2).
        - Each group's columns are sorted by numeric suffix (int).
        Returns: {prefix: [col_names_sorted]}.
        """
        if min_group_size < 2:
            raise ValueError("min_group_size must be >= 2.")

        # 1) Collect by prefix with parsed numeric suffix
        groups: dict[str, list[(int, str)]] = {}
        for s in candidates:
            sp = OneHotDecoder._split_prefix_numeric_suffix(s)
            if sp is None:
                continue
            prefix, sfx = sp
            groups.setdefault(prefix, []).append((int(sfx), s))

        # 2) Filter and finalize
        out: dict[str, list[str]] = {}
        for prefix, pairs in groups.items():
            if len(pairs) < min_group_size:
                continue
            pairs.sort(key=lambda t: t[0])  # sort by numeric suffix
            if require_consecutive:
                suffixes = [num for num, _ in pairs]
                # no duplicates
                if len(suffixes) != len(set(suffixes)):
                    continue
                # strictly consecutive run with step=1
                start = suffixes[0]
                if any(suffixes[i] != start + i for i in range(len(suffixes))):
                    continue
            out[prefix] = [name for _, name in pairs]

        return out

    def _is_one_hot_encoded(series: pd.Series) -> bool:
        try:
            # All rows should be the same type
            if series.apply(lambda x: type(x)).nunique() != 1:
                return False

            # First, handle string representations of True/False
            series_copy = series.copy()
            if series_copy.dtype == "object" or series_copy.dtype == "string":
                # Convert string representations of boolean values to numeric
                series_copy = series_copy.astype(str).str.strip().str.lower()
                series_copy = series_copy.replace({"true": "1", "false": "0"})

            # Column contains only 0 and 1 (as strings or numbers or booleans)
            series_copy = series_copy.astype(float)
            if set(series_copy.unique()) != {0.0, 1.0}:
                return False

            series_copy = series_copy.astype(int)

            # Column doesn't contain any NaN, np.NaN, space, null, etc.
            if not (series_copy.isin([0, 1])).all():
                return False

            vc = series_copy.value_counts()
            # Column should contain both 0 and 1
            if len(vc) != 2:
                return False

            # Minority class is 1
            if vc[1] >= vc[0]:
                return False

            return True
        except ValueError:
            return False



import pandas as pd
from pandas.api.types import (
    is_float_dtype,
    is_object_dtype,
    is_string_dtype,
)
import re

from upgini.utils.base_search_key_detector import BaseSearchKeyDetector


class PostalCodeSearchKeyDetector(BaseSearchKeyDetector):
    postal_pattern = re.compile(r'^[A-Za-z0-9][A-Za-z0-9\s\-]{1,9}$')

    def _is_search_key_by_name(self, column_name: str) -> bool:
        return "zip" in str(column_name).lower() or "postal" in str(column_name).lower()

    def _is_search_key_by_values(self, column: pd.Series) -> bool:
        """
        # Fast two-step check whether the column looks like a postal code.
        # Returns True if, after removing missing values, values remain,
        # and all of them match the common characteristics of a postal code.
        """
        # Check only columns that are candidates for postal code by column name
        if not self._is_search_key_by_name(column.name):
            return False

        s = column.copy().dropna().astype(str).str.strip()
        s = s[s != ""]  # remove empty strings
        if s.empty:
            return False

        # remove suffix ".0" (often after float)
        s = s.str.replace(r"\.0$", "", regex=True)

        # --- Step 1: fast filtering ---
        mask_len = s.str.len().between(2, 10)
        mask_digit = s.str.contains(r'\d', regex=True)
        mask_chars = ~s.str.contains(r'[^A-Za-z0-9\s\-]', regex=True)
        fast_mask = mask_len & mask_digit & mask_chars

        # if any of them failed the fast check, return False
        if not fast_mask.all():
            return False

        # --- Step 2: regex check ---
        # only if the first step passed
        valid_mask = s.apply(lambda x: bool(self.postal_pattern.fullmatch(x)))
        return valid_mask.all()


class PostalCodeSearchKeyConverter:

    def __init__(self, postal_code_column: str):
        self.postal_code_column = postal_code_column

    def convert(self, df: pd.DataFrame) -> pd.DataFrame:
        if is_string_dtype(df[self.postal_code_column]) or is_object_dtype(df[self.postal_code_column]):
            try:
                df[self.postal_code_column] = (
                    df[self.postal_code_column].astype("string").astype("float64").astype("Int64").astype("string")
                )
            except Exception:
                pass
        elif is_float_dtype(df[self.postal_code_column]):
            df[self.postal_code_column] = df[self.postal_code_column].astype("Int64").astype("string")

        df[self.postal_code_column] = (
            df[self.postal_code_column]
            .astype("string")
            .str.upper()
            .str.replace(r"[^0-9A-Z]", "", regex=True)  # remove non alphanumeric characters
            .str.replace(r"^0+\B", "", regex=True)  # remove leading zeros
        )
        # if (df[self.postal_code_column] == "").all():
        #     raise ValidationError(self.bundle.get("invalid_postal_code").format(self.postal_code_column))

        return df



import logging
from logging import Logger
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_float_dtype,
    is_integer_dtype,
    is_object_dtype,
    is_string_dtype,
)

from upgini.resource_bundle import bundle


class FeaturesValidator:
    def __init__(self, logger: Optional[Logger] = None):
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger()
            self.logger.setLevel("FATAL")

    def validate(
        self,
        df: pd.DataFrame,
        features: List[str],
        features_for_generate: Optional[List[str]] = None,
        columns_renaming: Optional[Dict[str, str]] = None,
    ) -> Tuple[List[str], List[str]]:
        empty_or_constant_features = []
        high_cardinality_features = []
        warnings = []

        for f in features:
            column = df[f]
            if is_object_dtype(column):
                column = column.astype("string")
            value_counts = column.value_counts(dropna=False, normalize=True)
            most_frequent_percent = value_counts.iloc[0]

            if len(value_counts) == 1:
                empty_or_constant_features.append(f)
            elif most_frequent_percent >= 0.99 and len(value_counts) != 2 and not self.__is_integer(column):
                empty_or_constant_features.append(f)

        columns_renaming = columns_renaming or {}

        if empty_or_constant_features:
            msg = bundle.get("empty_or_contant_features").format(
                [columns_renaming.get(f, f) for f in empty_or_constant_features]
            )
            warnings.append(msg)

        high_cardinality_features = self.find_high_cardinality(df[features])
        if features_for_generate:
            high_cardinality_features = [
                f for f in high_cardinality_features if columns_renaming.get(f, f) not in features_for_generate
            ]
        if high_cardinality_features:
            msg = bundle.get("high_cardinality_features").format(
                [columns_renaming.get(f, f) for f in high_cardinality_features]
            )
            warnings.append(msg)

        return (empty_or_constant_features + high_cardinality_features, warnings)

    @staticmethod
    def find_high_cardinality(df: pd.DataFrame) -> List[str]:
        # Remove high cardinality columns
        row_count = df.shape[0]
        if row_count < 100:  # For tests with small datasets
            return []
        return [
            i
            for i in df
            if (is_object_dtype(df[i]) or is_string_dtype(df[i]) or FeaturesValidator.__is_integer(df[i]))
            and (df[i].nunique(dropna=False) / row_count >= 0.85)
        ]

    @staticmethod
    def __is_integer(series: pd.Series) -> bool:
        return (
            is_integer_dtype(series)
            or series.dropna()
            .apply(
                lambda f: (float.is_integer(f) and abs(f) < np.iinfo(np.int64).max) if isinstance(f, float) else False
            )
            .all()
        )

    @staticmethod
    def find_constant_features(df: pd.DataFrame) -> List[str]:
        return [i for i in df if df[i].nunique() <= 1]


def is_effective_int(column: pd.Series) -> bool:
    """
    Return True if the series is effectively integer:
    - integer dtype; or
    - float dtype with all non-null values being whole numbers within int64 range.
    """
    if is_integer_dtype(column):
        return True
    if is_float_dtype(column):
        non_null_values = column.dropna()
        if non_null_values.empty:
            return False
        within_range_mask = non_null_values.abs() < np.iinfo(np.int64).max
        is_integer_mask = non_null_values.apply(lambda v: float(v).is_integer())
        return bool((within_range_mask & is_integer_mask).all())
    return False



import datetime
import logging
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from pandas.api.types import is_numeric_dtype

from upgini.errors import ValidationError
from upgini.metadata import CURRENT_DATE_COL, EVAL_SET_INDEX, SearchKey
from upgini.resource_bundle import ResourceBundle, get_custom_bundle
from upgini.utils.base_search_key_detector import BaseSearchKeyDetector

DATE_FORMATS = [
    "%Y-%m-%d",
    "%d.%m.%y",
    "%d.%m.%Y",
    "%m.%d.%y",
    "%m.%d.%Y",
    "%Y/%m/%d",
    "%y/%m/%d",
    "%d/%m/%Y",
    "%d/%m/%y",
    "%m/%d/%Y",
    "%m/%d/%y",
    "%Y-%m-%dT%H:%M:%S.%f",
]

DATETIME_PATTERN = r"^[\d\s\.\-:T/+]+$"


class DateSearchKeyDetector(BaseSearchKeyDetector):
    def _is_search_key_by_name(self, column_name: str) -> bool:
        lower_column_name = str(column_name).lower()
        return "date" in lower_column_name or "time" in lower_column_name or "timestamp" in lower_column_name

    def _is_search_key_by_values(self, column: pd.Series) -> bool:
        return DateTimeConverter(column.name).is_datetime(column.to_frame(column.name))


class DateTimeConverter:
    DATETIME_COL = "_date_time"
    # MIN_SUPPORTED_DATE_TS = datetime.datetime(1999, 12, 31)  # 946684800000  # 2000-01-01
    MIN_SUPPORTED_DATE_TS = pd.to_datetime(datetime.datetime(1999, 12, 31)).tz_localize(None)

    def __init__(
        self,
        date_column: str,
        date_format: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
        bundle: Optional[ResourceBundle] = None,
        generate_cyclical_features: bool = True,
    ):
        self.date_column = date_column
        self.date_format = date_format
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger("muted_logger")
            self.logger.setLevel("FATAL")
        self.generated_features: List[str] = []
        self.bundle = bundle or get_custom_bundle()
        self.generate_cyclical_features = generate_cyclical_features
        self.has_old_dates = False

    @staticmethod
    def _int_to_opt(i: int) -> Optional[int]:
        if i == -9223372036855:
            return None
        else:
            return i

    @staticmethod
    def clean_date(s: Optional[str]):
        try:
            if s is None or len(str(s).strip()) == 0:
                return None
            if sum(ch.isdigit() for ch in str(s)) < 6:
                return None
            return s
        except Exception:
            return None

    def is_datetime(self, df: pd.DataFrame) -> bool:
        if len(df) == 0 or df[self.date_column].isna().all():
            return False

        if pd.api.types.is_datetime64_any_dtype(df[self.date_column]):
            return True

        parsed = self.parse_datetime(df, raise_errors=False)
        return parsed is not None and parsed.isna().mean() <= 0.5

    def parse_datetime(self, df: pd.DataFrame, raise_errors=True) -> pd.Series | None:
        if len(df) == 0 or df[self.date_column].isna().all():
            return None

        date_col = df[self.date_column].copy()

        try:
            if date_col.apply(lambda x: isinstance(x, datetime.datetime)).all():
                parsed_datetime = date_col.apply(lambda x: x.replace(tzinfo=None))
            elif isinstance(date_col.dropna().values[0], datetime.date):
                parsed_datetime = pd.to_datetime(date_col, errors="coerce")
            elif isinstance(date_col.dtype, pd.PeriodDtype):
                parsed_datetime = date_col.dt.to_timestamp()
            elif is_numeric_dtype(date_col):
                # 315532801 - 2524608001    - seconds
                # 315532801000 - 2524608001000 - milliseconds
                # 315532801000000 - 2524608001000000 - microseconds
                # 315532801000000000 - 2524608001000000000 - nanoseconds
                if date_col.apply(lambda x: 10**16 < x).all():
                    parsed_datetime = pd.to_datetime(date_col, unit="ns")
                elif date_col.apply(lambda x: 10**14 < x < 10**16).all():
                    parsed_datetime = pd.to_datetime(date_col, unit="us")
                elif date_col.apply(lambda x: 10**11 < x < 10**14).all():
                    parsed_datetime = pd.to_datetime(date_col, unit="ms")
                elif date_col.apply(lambda x: 10**8 < x < 10**11).all():
                    parsed_datetime = pd.to_datetime(date_col, unit="s")
                else:
                    msg = self.bundle.get("unsupported_date_type").format(self.date_column)
                    if raise_errors:
                        raise ValidationError(msg)
                    else:
                        return None
            else:
                date_col = date_col.astype("string").apply(self.clean_date)
                parsed_datetime = self.parse_string_date(date_col.to_frame(self.date_column), raise_errors)
                if parsed_datetime.isna().all():
                    raise ValidationError(self.bundle.get("invalid_date_format").format(self.date_column))
            parsed_datetime = parsed_datetime.dt.tz_localize(None)
            return parsed_datetime
        except Exception as e:
            if raise_errors:
                raise ValidationError(e)
            else:
                return None

    def to_date_string(self, df: pd.DataFrame) -> pd.Series:
        parsed_datetime = self.parse_datetime(df)
        if parsed_datetime is None:
            return df[self.date_column]
        return parsed_datetime.dt.strftime("%Y-%m-%d")

    def to_date_ms(self, df: pd.DataFrame) -> pd.Series:
        parsed_datetime = self.parse_datetime(df)
        if parsed_datetime is None:
            return df[self.date_column]
        return self.convert_datetime_to_date_ms(parsed_datetime)

    def convert_datetime_to_datetime_ms(self, date_col: pd.Series) -> pd.Series:
        if date_col.dt.unit == "ns":
            date_col = date_col.astype(np.int64) // 1_000_000
        elif date_col.dt.unit == "us":
            date_col = date_col.astype(np.int64) // 1_000
        elif date_col.dt.unit == "ms":
            date_col = date_col.astype(np.int64)
        elif date_col.dt.unit == "s":
            date_col = date_col.astype(np.int64) * 1_000
        else:
            raise ValueError(f"Unsupported date unit: {date_col.dt.unit}")

        return date_col.apply(self._int_to_opt).astype("Int64")

    def convert_datetime_to_date_ms(self, date_col: pd.Series) -> pd.Series:
        date_col = date_col.dt.floor("D")
        return self.convert_datetime_to_datetime_ms(date_col)

    def convert(self, df: pd.DataFrame, keep_time=False) -> pd.DataFrame:
        df = df.copy()
        parsed_datetime = self.parse_datetime(df)
        if parsed_datetime is None:
            return df

        df[self.date_column] = parsed_datetime

        # If column with date is datetime then extract seconds of the day and minute of the hour
        # as additional features
        seconds = "datetime_seconds"

        df = self.clean_old_dates(df)

        # Define function to apply sine and cosine transformations
        def add_cyclical_features(df, column, period):
            period_suffix = f"_{period}" if column != "day_in_quarter" else ""
            sin_feature = f"datetime_{column}_sin{period_suffix}"
            cos_feature = f"datetime_{column}_cos{period_suffix}"
            if sin_feature not in df.columns:
                df[sin_feature] = np.sin(2 * np.pi * df[column] / period)
                self.generated_features.append(sin_feature)
            if cos_feature not in df.columns:
                df[cos_feature] = np.cos(2 * np.pi * df[column] / period)
                self.generated_features.append(cos_feature)

        if self.generate_cyclical_features:

            df["quarter"] = df[self.date_column].dt.quarter

            # Calculate the start date of the quarter for each timestamp
            df["quarter_start"] = df[self.date_column].dt.to_period("Q").dt.start_time

            # Calculate the day in the quarter
            df["day_in_quarter"] = (df[self.date_column] - df["quarter_start"]).dt.days + 1

            # Vectorized calculation of days_in_quarter
            quarter = df["quarter"]
            start = df["quarter_start"]
            year = start.dt.year
            month = start.dt.month

            quarter_end_year = np.where(quarter == 4, year + 1, year)
            quarter_end_month = np.where(quarter == 4, 1, month + 3)

            end = pd.to_datetime({"year": quarter_end_year, "month": quarter_end_month, "day": 1})
            end.index = df.index

            df["days_in_quarter"] = (end - start).dt.days

            add_cyclical_features(df, "day_in_quarter", df["days_in_quarter"])  # Days in the quarter

            df.drop(columns=["quarter", "quarter_start", "day_in_quarter", "days_in_quarter"], inplace=True)

            df[seconds] = (df[self.date_column] - df[self.date_column].dt.floor("D")).dt.seconds

            seconds_without_na = df[seconds].dropna()
            if (seconds_without_na != 0).any() and seconds_without_na.nunique() > 1:
                self.logger.info("Time found in date search key. Add extra features based on time")

                # Extract basic components
                df["second"] = df[self.date_column].dt.second
                df["minute"] = df[self.date_column].dt.minute
                df["hour"] = df[self.date_column].dt.hour

                # Apply cyclical transformations
                add_cyclical_features(df, "second", 60)  # Seconds in a minute
                add_cyclical_features(df, "minute", 60)  # Minutes in an hour
                add_cyclical_features(df, "minute", 30)  # Minutes in half an hour
                add_cyclical_features(df, "hour", 24)  # Hours in a day

                # Drop intermediate columns if not needed
                df.drop(columns=["second", "minute", "hour"], inplace=True)
            else:
                keep_time = False

            for generated_feature in self.generated_features[:]:
                if df[generated_feature].dropna().nunique() <= 1:
                    self.logger.warning(f"Generated constant feature {generated_feature} will be dropped")
                    df.drop(columns=generated_feature, inplace=True)
                    self.generated_features.remove(generated_feature)

            df.drop(columns=seconds, inplace=True)

        if keep_time:
            df[self.DATETIME_COL] = self.convert_datetime_to_datetime_ms(df[self.date_column])
        df[self.date_column] = self.convert_datetime_to_date_ms(df[self.date_column])

        self.logger.info(f"Date after convertion to timestamp: {df[self.date_column]}")

        return df

    def parse_string_date(self, df: pd.DataFrame, raise_errors=True) -> pd.Series | None:
        if self.date_format is not None:
            try:
                return pd.to_datetime(df[self.date_column], format=self.date_format)
            except ValueError as e:
                if raise_errors:
                    raise ValidationError(e)
                else:
                    return None
        else:
            for date_format in DATE_FORMATS:
                try:
                    return pd.to_datetime(df[self.date_column], format=date_format)
                except ValueError:
                    pass
            try:
                # Suppress warning for intentional fallback to dateutil parsing
                import warnings

                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", message="Could not infer format")
                    return pd.to_datetime(df[self.date_column])
            except ValueError:
                try:
                    return pd.to_datetime(df[self.date_column], format="mixed", errors="raise")
                except ValueError:
                    if raise_errors:
                        raise ValidationError(self.bundle.get("invalid_date_format").format(self.date_column))
                    else:
                        return None

    def clean_old_dates(self, df: pd.DataFrame) -> pd.DataFrame:
        condition = df[self.date_column] <= self.MIN_SUPPORTED_DATE_TS
        old_subset = df[condition]
        if len(old_subset) > 0:
            self.has_old_dates = True
            df.loc[condition, self.date_column] = None
            self.logger.info(f"Set to None: {len(old_subset)} of {len(df)} rows because they are before 2000-01-01")
        return df


def is_time_series(df: pd.DataFrame, date_col: str) -> bool:
    try:
        if df[date_col].isnull().any():
            return False

        df = pd.to_datetime(df[date_col]).to_frame()

        def rel(row):
            if not pd.isnull(row[date_col]) and not pd.isnull(row["shifted_date"]):
                return relativedelta(row[date_col], row["shifted_date"])

        value_counts = df[date_col].value_counts()
        # count with each date is constant
        if value_counts.nunique() == 1:
            # Univariate timeseries
            if value_counts.unique()[0] == 1:
                df["shifted_date"] = df[date_col].shift(1)
                # if dates cover full interval without gaps
                return df.apply(rel, axis=1).nunique() == 1

            # Multivariate timeseries
            df_with_unique_dates = df.drop_duplicates().copy()

            df_with_unique_dates["shifted_date"] = df_with_unique_dates[date_col].shift(1)
            # if unique dates cover full interval without gaps
            return df_with_unique_dates.apply(rel, axis=1).nunique() == 1

        return False
    except Exception:
        return False


def is_blocked_time_series(df: pd.DataFrame, date_col: str, search_keys: List[str]) -> bool:
    try:
        df = df.copy()
        seconds = "datetime_seconds"
        if isinstance(df[date_col].dtype, pd.PeriodDtype):
            df[date_col] = df[date_col].dt.to_timestamp()
        elif is_numeric_dtype(df[date_col]):
            df[date_col] = pd.to_datetime(df[date_col], unit="ms")
        else:
            df[date_col] = pd.to_datetime(df[date_col])
        df[date_col] = df[date_col].dt.tz_localize(None)
        df[seconds] = (df[date_col] - df[date_col].dt.floor("D")).dt.seconds

        seconds_without_na = df[seconds].dropna()
        columns_to_drop = [c for c in search_keys if c != date_col] + [seconds]
        df.drop(columns=columns_to_drop, inplace=True)
        # Date, not datetime
        if (seconds_without_na != 0).any() and seconds_without_na.nunique() > 1:
            return False

        nunique_dates = df[date_col].nunique()
        # Unique dates count more than 270
        if nunique_dates < 270:
            return False

        min_date = df[date_col].min()
        max_date = df[date_col].max()
        days_delta = (max_date - min_date).days + 1
        # Missing dates less than 30% (unique dates count and days delta between earliest and latest dates)
        if nunique_dates / days_delta < 0.3:
            return False

        accumulated_changing_columns = set()

        def check_differences(group: pd.DataFrame):
            changing_columns = group.columns[group.nunique(dropna=False) > 1].to_list()
            accumulated_changing_columns.update(changing_columns)

        def is_multiple_rows(group: pd.DataFrame) -> bool:
            return group.shape[0] > 1

        grouped = df.groupby(date_col)[[c for c in df.columns if c != date_col]]
        dates_with_multiple_rows = grouped.apply(is_multiple_rows).sum()

        # share of dates with more than one record is more than 99%
        if dates_with_multiple_rows / nunique_dates < 0.99:
            return False

        if df.shape[1] <= 3:
            return True

        grouped.apply(check_differences)
        return len(accumulated_changing_columns) <= 2
    except Exception:
        return False


def is_dates_distribution_valid(
    df: pd.DataFrame,
    search_keys: Dict[str, SearchKey],
) -> bool:
    try:
        maybe_date_col = SearchKey.find_key(search_keys, [SearchKey.DATE, SearchKey.DATETIME])

        if EVAL_SET_INDEX in df.columns:
            X = df.query(f"{EVAL_SET_INDEX} == 0")
        else:
            X = df

        if maybe_date_col is None:
            for col in X.columns:
                if col in search_keys:
                    continue
                try:
                    if isinstance(X[col].dtype, pd.PeriodDtype):
                        pass
                    elif pd.__version__ >= "2.0.0":
                        # Format mixed to avoid massive warnings
                        pd.to_datetime(X[col], format="mixed")
                    else:
                        pd.to_datetime(X[col])
                    maybe_date_col = col
                    break
                except Exception:
                    pass

        if maybe_date_col is None or maybe_date_col == CURRENT_DATE_COL:
            return True

        # Don't check if date column is constant
        if X[maybe_date_col].nunique() <= 1:
            return True

        if isinstance(X[maybe_date_col].dtype, pd.PeriodDtype):
            dates = X[maybe_date_col].dt.to_timestamp().dt.date
        elif pd.__version__ >= "2.0.0":
            dates = pd.to_datetime(X[maybe_date_col], format="mixed").dt.date
        else:
            dates = pd.to_datetime(X[maybe_date_col]).dt.date

        date_counts = dates.value_counts().sort_index()

        date_counts_1 = date_counts[: round(len(date_counts) / 2)]
        date_counts_2 = date_counts[round(len(date_counts) / 2) :]
        ratio = date_counts_2.mean() / date_counts_1.mean()

        return ratio >= 0.8 and ratio <= 1.2
    except Exception:
        return False



import hashlib
import os
import platform
import shutil
import subprocess
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd


def file_hash(path: str | os.PathLike, algo: str = "sha256") -> str:
    """
    Returns file hash using system utilities, working consistently on Windows/macOS/Linux.
    If no suitable utility is found, gracefully falls back to hashlib.

    Supported algo values (depend on OS and available utilities):
      - "md5", "sha1", "sha224", "sha256", "sha384", "sha512"
    On Windows uses `certutil`.
    On Linux uses `sha*sum` (e.g., sha256sum) or `shasum -a N`.
    On macOS uses `shasum -a N` or `md5` for MD5.
    """
    p = str(Path(path))

    sysname = platform.system().lower()
    algo = algo.lower()

    # -------- command attempts depending on OS --------
    candidates: list[list[str]] = []

    if sysname == "windows":
        # certutil supports: MD5, SHA1, SHA256, SHA384, SHA512
        name_map = {
            "md5": "MD5",
            "sha1": "SHA1",
            "sha224": None,  # certutil doesn't support
            "sha256": "SHA256",
            "sha384": "SHA384",
            "sha512": "SHA512",
        }
        cert_name = name_map.get(algo)
        if cert_name:
            candidates.append(["certutil", "-hashfile", p, cert_name])
    else:
        # Unix-like systems
        # 1) specialized *sum utility if available (usually present on Linux)
        sum_cmd = f"{algo}sum"  # md5sum, sha256sum, etc.
        if shutil.which(sum_cmd):
            candidates.append([sum_cmd, p])

        # 2) universal shasum with -a parameter (available on macOS and often on Linux)
        shasum_bits = {
            "sha1": "1",
            "sha224": "224",
            "sha256": "256",
            "sha384": "384",
            "sha512": "512",
        }
        if algo in shasum_bits and shutil.which("shasum"):
            candidates.append(["shasum", "-a", shasum_bits[algo], p])

        # 3) for MD5 on macOS there's often a separate `md5` utility
        if algo == "md5" and shutil.which("md5"):
            candidates.append(["md5", p])

    # -------- try system utilities --------
    for cmd in candidates:
        try:
            out = subprocess.check_output(cmd, text=True, stderr=subprocess.STDOUT)
            digest = _parse_hash_output(out, cmd[0])
            if digest:
                return digest.lower()
        except (subprocess.CalledProcessError, FileNotFoundError):
            continue  # try next candidate

    # -------- reliable fallback to hashlib --------
    import hashlib

    try:
        h = getattr(hashlib, algo)
    except AttributeError:
        raise ValueError(f"Algorithm not supported: {algo}")

    hasher = h()
    with open(p, "rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            hasher.update(chunk)
    return hasher.hexdigest().lower()


def _parse_hash_output(output: str, tool: str) -> Optional[str]:
    """
    Converts output from different utilities to clean hash.
    Supports:
      - sha*sum / shasum: '<hex>  <filename>'
      - certutil (Windows): line with second element as hash (spaces inside are removed)
      - md5 (macOS): 'MD5 (file) = <hex>'
    """
    tool = tool.lower()
    lines = [ln.strip() for ln in output.splitlines() if ln.strip()]

    if not lines:
        return None

    if tool in {"sha1sum", "sha224sum", "sha256sum", "sha384sum", "sha512sum", "md5sum", "shasum"}:
        # format: '<hex>  <filename>'
        first = lines[0]
        parts = first.split()
        return parts[0] if parts else None

    if tool == "certutil":
        # format:
        # SHA256 hash of file <path>:
        # <AA BB CC ...>
        # CertUtil: -hashfile command completed successfully.
        if len(lines) >= 2:
            # Second line contains hex with spaces
            candidate = lines[1].replace(" ", "")
            # ensure it's hex
            if all(c in "0123456789abcdefABCDEF" for c in candidate):
                return candidate
        return None

    if tool == "md5":
        # format: 'MD5 (<file>) = <hex>'
        last = lines[-1]
        if "=" in last:
            return last.split("=", 1)[1].strip()
        # sometimes md5 can return just the hash
        parts = last.split()
        if parts and all(c in "0123456789abcdefABCDEF" for c in parts[-1]):
            return parts[-1]
        return None

    # as a last resort: take the first "looks like hash" word
    for ln in lines:
        for token in ln.split():
            if all(c in "0123456789abcdefABCDEF" for c in token) and len(token) >= 32:
                return token
    return None


def hash_input(X: pd.DataFrame, y: Optional[pd.Series] = None, eval_set: Optional[List[Tuple]] = None) -> str:
    hashed_objects = []
    try:
        hashed_objects.append(pd.util.hash_pandas_object(X, index=False).values)
        if y is not None:
            hashed_objects.append(pd.util.hash_pandas_object(y, index=False).values)
        if eval_set is not None:
            if isinstance(eval_set, tuple):
                eval_set = [eval_set]
            for eval_X, eval_y in eval_set:
                hashed_objects.append(pd.util.hash_pandas_object(eval_X, index=False).values)
                hashed_objects.append(pd.util.hash_pandas_object(eval_y, index=False).values)
        common_hash = hashlib.sha256(np.concatenate(hashed_objects)).hexdigest()
        return common_hash
    except Exception:
        return ""
